unit IDEExplorerForm;

{                                                                           }
{ File:       IDEExplorerForm.pas                                           }
{ Function:   Main form for IDE Explorer                                    }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2005 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

{$IF RTLVersion < 17.0}
  {$MESSAGE ERROR 'Only in Delphi 2005 or higher'}
{$IFEND}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, ExtCtrls, Menus, ImgList, TypInfo;

type
  TSearchType = (stName, stType);

  TObjectData = class
  public
    Obj: TObject;
    Level: Integer;
  end;

  TExplorerForm = class(TForm)
    OwnedButton: TButton;
    TopPanel: TPanel;
    ExplorerTree: TTreeView;
    ExplorerSplitter: TSplitter;
    ParentedButton: TButton;
    TreePopupMenu: TPopupMenu;
    ExpandAllItem: TMenuItem;
    CollapseAllItem: TMenuItem;
    CloseButton: TButton;
    AboutButton: TButton;
    PageControl: TPageControl;
    PropertiesPage: TTabSheet;
    TreesPage: TTabSheet;
    PropertiesList: TListView;
    TreesTree: TTreeView;
    NameSearchButton: TButton;
    NextSearchButton: TButton;
    StateImageList: TImageList;
    SmallImages: TImageList;
    procedure PropertiesListMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormShow(Sender: TObject);
    procedure NextSearchButtonClick(Sender: TObject);
    procedure SearchButtonClick(Sender: TObject);
    procedure TreesTreeCustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode;
      State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure AboutButtonClick(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure CollapseAllItemClick(Sender: TObject);
    procedure ExpandAllItemClick(Sender: TObject);
    procedure ParentedButtonClick(Sender: TObject);
    procedure ExplorerTreeChange(Sender: TObject; Node: TTreeNode);
    procedure OwnedButtonClick(Sender: TObject);
    procedure ExplorerTreeEditing(Sender: TObject; Node: TTreeNode;
      var AllowEdit: Boolean);
  private
    FSearchText: string;
    FSearchType: TSearchType;
    FLastButton: TObject;
    procedure RepeatLastAction;
    procedure SearchItem(Node: TTreeNode; Text: string;
      SearchType: TSearchType);
    procedure UpdateTreesTree(Node: TTreeNode);
    procedure FillTreeParented;
    procedure UpdateTreeParented;
    procedure FillTreeOwned;
    procedure UpdateTreeOwned;
    procedure UpdateListView(Node: TTreeNode);
    procedure ListProperties(Obj: TObject; Index, Level: Integer);
    procedure ToggleSubProperties(Item: TListItem);
    procedure UnlistProperties(Index: Integer; Obj: TObject);
    procedure InsertProperty(Obj: TObject; P: PPropList; I: Integer; Index: Integer; Level: Integer);
  public

  end;

var
  ExplorerForm: TExplorerForm;

implementation

uses
  IDEEXplorerAboutForm, IDEExplorerSearchForm;

resourcestring
  StrInheritance = 'Inheritance:';
  StrOwnership = 'Ownership:';
  StrParentage = 'Parentage:';
  StrSearchStringNotFound = 'Search string ''%s'' not found';
  StrPropsInaccessible = 'Properties inaccessible';
  StrUnnamed = '<unnamed>';

  {$R *.dfm}

function ItemStr(Comp: TComponent): string;
var
  Name: string;
begin
  Name := Comp.Name;
  if Name = '' then
    Name := StrUnnamed;
  Result := Format('%s: %s', [Name, Comp.ClassName]);
end;

procedure TExplorerForm.FillTreeOwned;

  procedure EnumComponents(Node: TTreeNode; ParentComp: TComponent);
  var
    NewNode: TTreeNode;
    Comp: TComponent;
  begin
    for Comp in ParentComp do
    begin
      NewNode := ExplorerTree.Items.AddChildObject(Node, ItemStr(Comp), Comp);
      EnumComponents(NewNode, Comp);
    end;
  end;

var
  AppNode: TTreeNode;
begin
  AppNode := ExplorerTree.Items.AddObjectFirst(nil, ItemStr(Application),
    Application);
  EnumComponents(AppNode, Application);
end;

procedure TExplorerForm.FillTreeParented;

  procedure EnumComponents(Node: TTreeNode; ParentControl: TWinControl);
  var
    NewNode: TTreeNode;
    Ctrl: TControl;
    I: Integer;
  begin
    for I := 0 to ParentControl.ControlCount - 1 do
    begin
      Ctrl := ParentControl.Controls[I];
      NewNode := ExplorerTree.Items.AddChildObject(Node,
        ItemStr(Ctrl), Ctrl);
      if Ctrl is TWinControl then
        EnumComponents(NewNode, TWinControl(Ctrl));
    end;
  end;

var
  AppNode: TTreeNode;
  F: TCustomForm;
  I: Integer;
begin
  for I := 0 to Screen.CustomFormCount - 1 do
  begin
    F := Screen.CustomForms[I];
    if not Assigned(F.Parent) then
    begin
      AppNode := ExplorerTree.Items.AddObjectFirst(nil, ItemStr(F), F);
      EnumComponents(AppNode, F);
    end;
  end;
end;

procedure TExplorerForm.UpdateTreeOwned;
var
  OldCursor: TCursor;
begin
  OldCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    ExplorerTree.Items.BeginUpdate;
    try
      ExplorerTree.Items.Clear;
      ExplorerTree.SortType := stNone;
      FillTreeOwned;
    finally
      ExplorerTree.Items.EndUpdate;
      ExplorerTree.SortType := stText;
    end;
  finally
    Screen.Cursor := OldCursor;
  end;
end;

procedure TExplorerForm.UpdateTreeParented;
var
  OldCursor: TCursor;
begin
  OldCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    ExplorerTree.Items.BeginUpdate;
    try
      ExplorerTree.Items.Clear;
      ExplorerTree.SortType := stNone;
      FillTreeParented;
    finally
      ExplorerTree.Items.EndUpdate;
      ExplorerTree.SortType := stText;
    end;
  finally
    Screen.Cursor := OldCursor;
  end;
end;


procedure TExplorerForm.OwnedButtonClick(Sender: TObject);
begin
  UpdateTreeOwned;
  FLastButton := Sender;
end;

procedure TExplorerForm.UpdateListView(Node: TTreeNode);
begin
  if Node = nil then Exit;
  PropertiesList.Items.BeginUpdate;
  try
    PropertiesList.Items.Clear;
    try
      ListProperties(TObject(Node.Data), -1, 0);
    except
      // error when property list unavailable (e.g. component in package)
      // simply ignore
    end;
  finally
    PropertiesList.Items.EndUpdate;
  end;
end;

procedure TExplorerForm.ListProperties(Obj: TObject; Index, Level: Integer);
var
  P: PPropList;
  I: Integer;
  NumProps: Integer;
begin
  NumProps := GetPropList(Obj, P);
  try
    for I := 0 to NumProps - 1 do
      InsertProperty(Obj, P, I, Index, Level);
  finally
    FreeMem(P);
  end;
end;

procedure TExplorerForm.ToggleSubProperties(Item: TListItem);
var
  Obj: TObject;
  Index: Integer;
begin
  Obj := TObject(Item.Data);
  Index := PropertiesList.Items.IndexOf(Item);
  if Assigned(Obj) then
  begin
    case Item.StateIndex of
      0:
        begin
          try
            ListProperties(Obj, Index + 1, Item.Indent + 1);
            Item.StateIndex := 1;
          except
            Item.StateIndex := -1;
          end;
        end;
      1:
        begin
          UnlistProperties(Index, Obj);
          Item.StateIndex := 0;
        end;
    end;
  end
  else
    Item.StateIndex := -1;
end;

procedure TExplorerForm.UnlistProperties(Index: Integer; Obj: TObject);
var
  I: Integer;
  P: PPropList;
  NumProps: Integer;
  Item: TListItem;
begin
  begin
    NumProps := GetPropList(Obj, P);
    PropertiesList.Items.BeginUpdate;
    try
      for I := 1 to NumProps do
      begin
        Item := PropertiesList.Items[Index + 1];
        if Assigned(Item.Data) and (Item.StateIndex = 1) then
          UnlistProperties(Index + 1, TObject(Item.Data));
        PropertiesList.Items.Delete(Index + 1);
      end;
    finally
      PropertiesList.Items.EndUpdate;
    end;
  end;
end;

procedure TExplorerForm.InsertProperty(Obj: TObject; P: PPropList; I: Integer; Index: Integer; Level: Integer);
var
  Item: TListItem;
  PropName: string;
  NewObj: TObject;
begin
  begin
    Item := PropertiesList.Items.Insert(Index + I);
    PropName := P^[I]^.Name;
    Item.Caption := PropName;
    Item.SubItems.Add(P^[I]^.PropType^.Name);
    Item.SubItems.Add(GetPropValue(Obj, PropName));
    Item.ImageIndex := Integer(PropType(Obj, PropName));
    Item.Indent := Level;
    if PropType(Obj, PropName) = tkClass then
    begin
      NewObj := GetObjectProp(Obj, PropName, TObject);
      Item.StateIndex := 0;
      Item.Data := NewObj;
    end;
  end;
end;

procedure TExplorerForm.UpdateTreesTree(Node: TTreeNode);

  // Recurses up until ParentClass is nil, adds nodes on the way back.

  function AddAncestor(T: TClass): TTreeNode;
  var
    ParentNode: TTreeNode;
  begin
    if T = nil then
    begin
      Result := TreesTree.Items.AddChild(nil, StrInheritance);
      Exit;
    end;
    ParentNode := AddAncestor(T.ClassParent);
    Result := TreesTree.Items.AddChild(ParentNode, T.ClassName);
    if Assigned(ParentNode) then
      ParentNode.Expand(False);
  end;

  // Recurses up until Owner is nil, adds nodes on the way back.

  function AddOwner(T: TComponent): TTreeNode;
  var
    ParentNode: TTreeNode;
  begin
    if T = nil then
    begin
      TreesTree.Items.AddChild(nil, ''); // empty row
      Result := TreesTree.Items.AddChild(nil, StrOwnership);
      Exit;
    end;
    ParentNode := AddOwner(T.Owner);
    Result := TreesTree.Items.AddChild(ParentNode, ItemStr(T));
    if Assigned(ParentNode) then
      ParentNode.Expand(False);
  end;

  // Recurses up until Parent is nil, adds nodes on the way back.

  function AddParent(T: TControl): TTreeNode;
  var
    ParentNode: TTreeNode;
  begin
    if T = nil then
    begin
      TreesTree.Items.AddChild(nil, '');
      Result := TreesTree.Items.AddChild(nil, StrParentage);
      Exit;
    end;
    ParentNode := AddParent(T.Parent);
    Result := TreesTree.Items.AddChild(ParentNode, ItemStr(T));
    if Assigned(ParentNode) then
      ParentNode.Expand(False);
  end;

var
  Comp: TComponent;

begin
  if Node = nil then Exit;
  TreesTree.Items.Clear;
  Comp := TObject(Node.Data) as TComponent;

  // Inheritance tree
  AddAncestor(Comp.ClassType);

  // Ownership tree
  AddOwner(Comp);

  // Parentage tree
  if Comp is TControl then
    AddParent(TControl(Comp));
end;

procedure TExplorerForm.ExplorerTreeEditing(Sender: TObject;
  Node: TTreeNode; var AllowEdit: Boolean);
begin
  AllowEdit := False;
  UpdateListView(Node);
end;

procedure TExplorerForm.ExplorerTreeChange(Sender: TObject; Node: TTreeNode);
begin
  UpdateListView(Node);
  UpdateTreesTree(Node);
end;

procedure TExplorerForm.ParentedButtonClick(Sender: TObject);
begin
  UpdateTreeParented;
  FLastButton := Sender;
end;

procedure TExplorerForm.ExpandAllItemClick(Sender: TObject);
begin
  ExplorerTree.FullExpand;
end;

procedure TExplorerForm.CollapseAllItemClick(Sender: TObject);
begin
  ExplorerTree.FullCollapse;
end;

procedure TExplorerForm.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

procedure TExplorerForm.AboutButtonClick(Sender: TObject);
var
  About: TIDEXAboutBox;
begin
  About := TIDEXAboutBox.Create(Self);
  try
    About.ShowModal;
  finally
    About.Free;
  end;
end;

procedure TExplorerForm.TreesTreeCustomDrawItem(Sender: TCustomTreeView;
  Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
begin
  if not Assigned(Node.Parent) then
    with Sender.Canvas.Font do
      Style := Style + [fsBold];
end;

procedure TExplorerForm.SearchItem(Node: TTreeNode; Text: string; SearchType:
  TSearchType);
var
  NodeText: string;
  Comp: TComponent;
begin
  while Assigned(Node) do
  begin
    Comp := TObject(Node.Data) as TComponent;
    case SearchType of
      stName: NodeText := Comp.Name;
      stType: NodeText := Comp.ClassName;
    end;
    if Pos(UpperCase(Text), UpperCase(NodeText)) > 0 then
    begin
      ExplorerTree.Select(Node);
      Exit;
    end;
    Node := Node.GetNext;
  end;
  ShowMessageFmt(StrSearchStringNotFound, [Text]);
end;

procedure TExplorerForm.SearchButtonClick(Sender: TObject);
var
  SearchForm: TSearchForm;
begin
  SearchForm := TSearchForm.Create(Self);
  try
    if IsPositiveResult(SearchForm.ShowModal) then
    begin
      if SearchForm.NameRadio.Checked then
        FSearchType := stName
      else
        FSearchType := stType;
      FSearchText := SearchForm.InputEdit.Text;
      SearchItem(ExplorerTree.Selected, FSearchText, FSearchType);
    end;
  finally
    SearchForm.Free;
  end;
end;

procedure TExplorerForm.NextSearchButtonClick(Sender: TObject);
begin
  if Assigned(ExplorerTree.Selected) then
    SearchItem(ExplorerTree.Selected.GetNext, FSearchText, FSearchType);
end;

procedure TExplorerForm.RepeatLastAction;
begin
  if FLastButton = OwnedButton then
    OwnedButtonClick(OwnedButton)
  else
    ParentedButtonClick(ParentedButton);
end;

procedure TExplorerForm.FormShow(Sender: TObject);
begin
  RepeatLastAction;
end;

procedure TExplorerForm.PropertiesListMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Item: TListItem;
  LeftX: Integer;
begin
  Item := PropertiesList.GetItemAt(X, Y);
  LeftX := Item.Indent * PropertiesList.SmallImages.Width;
  if (X >= LeftX) and (X < LeftX + PropertiesList.StateImages.Width) then
    ToggleSubProperties(Item);
end;

end.

